# This file is part of pyknos, a library for conditional density estimation.
# pyknos is licensed under the Affero General Public License v3,
# see <https://www.gnu.org/licenses/>.

VERSION = (0, 15, 1)

__version__ = ".".join(map(str, VERSION))
